==> unity-ads-4.12.4-sources.jar: com/unity3d/services/SDKErrorHandler.kt <==
package com.unity3d.services


import androidx.webkit.internal.ApiFeature.T
import com.unity3d.ads.core.configuration.AlternativeFlowReader
import com.unity3d.ads.core.domain.SendDiagnosticEvent
import com.unity3d.ads.core.domain.SendDiagnosticEvent.Companion.COROUTINE_NAME
import com.unity3d.ads.core.domain.SendDiagnosticEvent.Companion.REASON
import com.unity3d.ads.core.domain.SendDiagnosticEvent.Companion.REASON_DEBUG
import com.unity3d.ads.core.extensions.retrieveUnityCrashValue
import com.unity3d.services.UnityAdsConstants.ErrorHandler.ERROR_HANDLER_STACK_TRACE_LINES_MAX
import com.unity3d.services.core.log.DeviceLog
import com.unity3d.services.core.request.metrics.Metric
import com.unity3d.services.core.request.metrics.SDKMetricsSender
import kotlinx.coroutines.CoroutineDispatcher
import kotlinx.coroutines.CoroutineExceptionHandler
import kotlinx.coroutines.CoroutineName
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.launch
import kotlinx.coroutines.plus
import java.io.PrintWriter
import java.io.StringWriter
import kotlin.coroutines.CoroutineContext
import kotlin.math.min


class SDKErrorHandler(
    private val ioDispatcher: CoroutineDispatcher,
    private val alternativeFlowReader: AlternativeFlowReader,
    private val sendDiagnosticEvent: SendDiagnosticEvent,
    private val sdkMetricsSender: SDKMetricsSender
) : CoroutineExceptionHandler {
    private val scope = CoroutineScope(ioDispatcher) + CoroutineName("SDKErrorHandler")

    override val key = CoroutineExceptionHandler.Key

    override fun handleException(context: CoroutineContext, exception: Throwable) {
        val coroutineName = retrieveCoroutineName(context)
        val name: String = when (exception) {
            is NullPointerException -> "native_exception_npe"
            is OutOfMemoryError -> "native_exception_oom"
            is IllegalStateException -> "native_exception_ise"
            is SecurityException -> "native_exception_se"
            is RuntimeException -> "native_exception_re"
            else -> "native_exception"
        }

        val isAlternativeFlowEnabled = alternativeFlowReader()
        val crashValue = exception.retrieveUnityCrashValue()
        DeviceLog.error("Unity Ads SDK encountered an exception: $crashValue")
        if (isAlternativeFlowEnabled) {
            val stackTrace = getShortenedStackTrace(exception, ERROR_HANDLER_STACK_TRACE_LINES_MAX)
            sendDiagnostic(name, crashValue, coroutineName, stackTrace)
        } else {
            sendMetric(Metric(name, crashValue))
        }

    }

    private fun sendDiagnostic(name: String, reason: String, scopeName: String, stackTrace: String) {
        scope.launch {
            sendDiagnosticEvent(
                event = name,
                tags = mapOf(
                    REASON to reason,
                    REASON_DEBUG to stackTrace,
                    COROUTINE_NAME to scopeName
                )
            )
        }
    }

    private fun sendMetric(metric: Metric) = sdkMetricsSender.sendMetric(metric)

    private fun retrieveCoroutineName(context: CoroutineContext) = context[CoroutineName]?.name ?: "unknown"

    private fun getShortenedStackTrace(e: Throwable, maxLines: Int): String {
        try {
            val writer = StringWriter()
            e.printStackTrace(PrintWriter(writer))
            return writer.toString()
                .trim()
                .lineSequence()
                .take(maxLines)
                .joinToString("\n")
        } catch (t: Throwable) {
            return ""
        }
    }

    companion object {
        const val UNKNOWN_FILE = "unknown"
        const val UNITY_PACKAGE = "com.unity3d"
    }
}

==> unity-ads-4.16.3-sources.jar: com/unity3d/services/SDKErrorHandler.kt <==
package com.unity3d.services


import com.unity3d.ads.core.configuration.AlternativeFlowReader
import com.unity3d.ads.core.domain.SendDiagnosticEvent
import com.unity3d.ads.core.domain.SendDiagnosticEvent.Companion.COROUTINE_NAME
import com.unity3d.ads.core.domain.SendDiagnosticEvent.Companion.REASON
import com.unity3d.ads.core.domain.SendDiagnosticEvent.Companion.REASON_DEBUG
import com.unity3d.ads.core.extensions.getShortenedStackTrace
import com.unity3d.ads.core.extensions.retrieveUnityCrashValue
import com.unity3d.services.UnityAdsConstants.ErrorHandler.ERROR_HANDLER_STACK_TRACE_LINES_MAX
import com.unity3d.services.core.log.DeviceLog
import com.unity3d.services.core.request.metrics.Metric
import com.unity3d.services.core.request.metrics.SDKMetricsSender
import kotlinx.coroutines.CoroutineDispatcher
import kotlinx.coroutines.CoroutineExceptionHandler
import kotlinx.coroutines.CoroutineName
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.launch
import kotlinx.coroutines.plus
import kotlin.coroutines.CoroutineContext


class SDKErrorHandler(
    private val ioDispatcher: CoroutineDispatcher,
    private val alternativeFlowReader: AlternativeFlowReader,
    private val sendDiagnosticEvent: SendDiagnosticEvent,
    private val sdkMetricsSender: SDKMetricsSender
) : CoroutineExceptionHandler {
    private val scope = CoroutineScope(ioDispatcher) + CoroutineName("SDKErrorHandler")

    override val key = CoroutineExceptionHandler.Key

    override fun handleException(context: CoroutineContext, exception: Throwable) {
        val coroutineName = retrieveCoroutineName(context)
        val name: String = when (exception) {
            is NullPointerException -> "native_exception_npe"
            is OutOfMemoryError -> "native_exception_oom"
            is IllegalStateException -> "native_exception_ise"
            is SecurityException -> "native_exception_se"
            is RuntimeException -> "native_exception_re"
            else -> "native_exception"
        }

        val isAlternativeFlowEnabled = alternativeFlowReader()
        val crashValue = exception.retrieveUnityCrashValue()
        DeviceLog.error("Unity Ads SDK encountered an exception: $crashValue")
        if (isAlternativeFlowEnabled) {
            val stackTrace = exception.getShortenedStackTrace(ERROR_HANDLER_STACK_TRACE_LINES_MAX)
            sendDiagnostic(name, crashValue, coroutineName, stackTrace)
        } else {
            sendMetric(Metric(name, crashValue))
        }

    }

    private fun sendDiagnostic(name: String, reason: String, scopeName: String, stackTrace: String) {
        scope.launch {
            sendDiagnosticEvent(
                event = name,
                tags = mapOf(
                    REASON to reason,
                    REASON_DEBUG to stackTrace,
                    COROUTINE_NAME to scopeName
                )
            )
        }
    }

    private fun sendMetric(metric: Metric) = sdkMetricsSender.sendMetric(metric)

    private fun retrieveCoroutineName(context: CoroutineContext) = context[CoroutineName]?.name ?: "unknown"

    companion object {
        const val UNKNOWN_FILE = "unknown"
        const val UNITY_PACKAGE = "com.unity3d"
    }
}
